/*
 * SpeedDepthWindNMEA.30.ino, 20191026, PGM
 * Integrating the pin interrupt sketches for 
 *    Nexus speed, depth, and wind sensor measurements
 *    Serial3 for Nexus NMEA data
 *    Serial2 for GPS data
 *    Serial for Serial Monitor and HM10 BLE transceiver
 *
 * This sketch assumes that the Nexus's two depth sounder pulses (transmit and receive) have been buffered with an RS latch
 * reducing the two signals to one proportional to the ultrasonic pulse time-of-flight (ToF). The ToF, reported in 
 * microseconds, can be converted to water depth via Depth = ToF * 4800 / 2 where 4800 is speed of sound in water (ft/s).
 *
 * The algorithm for the wind measurements came from MeasurePulseWidthAndPeriodTwoEncoders.1.ino, 20181023, PGM
 * This sketch reads the two Nexus wind instrument quadrature sensors for their raw inputs into the AWS and AWA calculations. 
 * See .2.ino for AWA calculation.
 * 
 * The algorithm for NMEA logging came from LogGPSDataToSDCardMega.ino with the SD card functions deleted. 
 * SD card write operations take too long.
 *
 */

#include <EnableInterrupt.h>

#define alpha 0.2                        //low pass filter (EWMA) weighting function; should alpha be adapted to the SNR?

//Define interrupt pins for the Mega2560 (2, 3, 18-21, note that 20 and 21 are also I2C):
#define pinSpeed    18                   //hardware interrupt pin
#define pinDepth    19                   //hardware interrupt pin
#define pinWind1    2                    //hardware interrupt pin
#define pinWind2    3                    //hardware interrupt pin


volatile byte MeasurementComplete = 0;       //MeasurementComplete will take on the pin values, e.g. pinSpeed, ...; 0 is no interrupt
volatile unsigned long tSpeed = 0;           //Speed: time between CHANGE events, i.e. 1/2 of the period
volatile unsigned long tDepth = 0;           //Depth: sound pulse time-of-flight, i.e. down and back
volatile unsigned long tWind1 = 0;           //tWind1: positive pulse width of 1st wind quadrature channel
volatile unsigned long tWind2 = 0;
volatile unsigned long pWind = 0;            //Wind: time between RISING, FALLING, and RISING events, i.e. 1/2 of the period, used for both quadrature channels
volatile unsigned long dcWind1 = 0;          //Wind1: duty cycle of 1st wind quadrature channel
volatile unsigned long dcWind2 = 0;
volatile unsigned long t0[] = {0, 0, 0, 0};  //time zero of the rising edge of the pulse for Speed, Depth, Wind1, and Wind2

//Nexus server NMEA variables
String inputString3 = "";            // a String to hold incoming data
boolean stringComplete3 = false;     // whether the string is complete

//Garmin GPS variables
String inputString2 = "";            // a String to hold incoming data
boolean stringComplete2 = false;     // whether the string is complete 

void setup() 
    {
    Serial.begin(115200);        // Serial Monitor, may also have the HM10 BLE transceiver
    Serial2.begin(4800);        // Garmin GPS is 4800 baud, on Mega pin 17
    Serial3.begin(4800);        // Nexus NMEA 0183 is 4800 baud, on Mega pin 15
    inputString2.reserve(200);  // reserve 200 bytes for the Serial2 inputString
    inputString3.reserve(200);  // reserve 200 bytes for the Serial3 inputString    

    pinMode(pinSpeed,INPUT);    //Relying on Nexus's pullup resistors, not the Arduino's internal ones
    pinMode(pinDepth,INPUT);    //If implementing the hardware without the Nexus server these inputs will require pullup resistors
    pinMode(pinWind1,INPUT);
    pinMode(pinWind2,INPUT);  
  
    //Attach interrupts:
    enableInterrupt(pinSpeed, ISR_Speed_CHANGE, CHANGE);            //knotmeter pulse is symmetric, i.e. 50% duty cycle, so CHANGE is sufficient
    enableInterrupt(pinDepth, ISR_Depth_RISING, RISING);            //depth measurement starts on the rising edge of the positive pulse, ends on the falling
    enableInterrupt(pinWind1, ISR_Wind1_RISING, RISING);                
    enableInterrupt(pinWind2, ISR_Wind2_RISING, RISING);

    Serial.println("This is SpeedDepthWindNMEA.30.ino.");       //Identify the program in case I lose track of what's on the uC
    Serial.println("Initialization done. Logging data ...");
  
    }
 
 
void loop() 
    { 
    static long t0;        //update timer                                            //data update timer zero
        
    switch (MeasurementComplete)                    
        {
        case pinSpeed:
            //BoatSpeed = 0.278E6 / (2 * tSpeed)        
            break;
        case pinDepth:
            //Depth = tDepth * 4800 / 2 where tDepth and 4800 (speed of sound in water) are in seconds
            //Depth = tDepth * 0.0024        //tDepth in us
            break;
        case pinWind1:
            //http://forums.sailinganarchy.com/index.php?/topic/172637-old-nexus-silva-wind-instrument-problem/
            //AWS = (1E6 / pWind / 3.4 - 0.3) * 1.944
            //AWS = 5718 / pWind - 0.58
            break;
        case pinWind2:
        
            break;
        case 0:
            //No interrupt occurred
            break;
        }            //end of switch (MeasurementComplete)
    MeasurementComplete = 0;                                 //Reset the MeasurementComplete flag after the new data have been processed
    

    if ((millis() - t0) > 400)
        {
        String strNexus = "$NEXUS";                  //build the Nexus sensors data string
        strNexus = strNexus + "," + tSpeed + "," + tDepth + "," + dcWind1 + "," + dcWind2 + "," + pWind;   //string of appended responses
        Serial.println(strNexus);
        t0 = millis();             //reset the timer
        }
    
    //Check progress on the Nexus NMEA data:
    if (stringComplete3) 
        {
        Serial.print(inputString3);        // echo the data to the serial monitor
        inputString3 = "";                 // clear the inputString:
        stringComplete3 = false;           // reset the Boolean for the next string
        }    

    //Check progress on the Garmin GPS data:
    if (stringComplete2) 
        {
        Serial.print(inputString2);        // echo the data to the serial monitor
        inputString2 = "";                 // clear the inputString:
        stringComplete2 = false;           // reset the Boolean for the next string
        }    
        
    }  //loop()
    
    
void ISR_Speed_CHANGE()
{
MeasurementComplete = pinSpeed;          //set MeasurementComplete value to process data in loop()
long now = micros();
//tSpeed = now - t0[0];                                     // raw/unfiltered time
tSpeed = (1 - alpha) * tSpeed + alpha * (now - t0[0]);      // low pass filtered time
t0[0] = now;
}

    
void ISR_Depth_RISING()
{
MeasurementComplete = 0;                 //in the middle of a measurement so clear the MeasurementComplete flag
t0[1] = micros();                                           //reset t0
enableInterrupt(pinDepth, ISR_Depth_FALLING, FALLING);
}


void ISR_Depth_FALLING()
{
MeasurementComplete = pinDepth;           //measurement complete, set the MeasurementComplete flag to process the data                                            
//tDepth = micros() - t0[1];
tDepth = (1 - alpha) * tDepth + alpha * (micros() - t0[1]);        
enableInterrupt(pinDepth, ISR_Depth_RISING, RISING);
}


void ISR_Wind1_RISING()
{
MeasurementComplete = pinWind1;           //measurement complete, set the MeasurementComplete flag to process the data
unsigned long now = micros();
//dcWind1 = 1E4 * tWind1 / (now - t0[2]);                   //Warning: Integer division will only return a value  from 0 to 1E4
dcWind1 = (1 - alpha) * dcWind1 + alpha * 1E4 * tWind1 / (now - t0[2]);    
//pWind = now - t0[2]    
pWind = (1 - alpha) * pWind + alpha * (now - t0[2]);        //pWind is updated for both quadrature channels
t0[2] = now;
enableInterrupt(pinWind1, ISR_Wind1_FALLING, FALLING);
}


void ISR_Wind1_FALLING()
{
MeasurementComplete = 0;                //in the middle of a measurement so clear the MeasurementComplete flag
tWind1 = micros() - t0[2];
enableInterrupt(pinWind1, ISR_Wind1_RISING, RISING);
}    
    
    
void ISR_Wind2_RISING()
{
MeasurementComplete = pinWind2;          //measurement complete, set the MeasurementComplete flag to process the data
unsigned long now = micros();
//dcWind2 = 1E4 * tWind2 / (now - t0[3]);
dcWind2 = (1 - alpha) * dcWind2 + alpha * 1E4 * tWind2 / (now - t0[3]);
//pwind = now - t0[3]
pWind = (1 - alpha) * pWind + alpha * (now - t0[3]);        
t0[3] = now;
enableInterrupt(pinWind2, ISR_Wind2_FALLING, FALLING);
}


void ISR_Wind2_FALLING()
{
MeasurementComplete = 0;                  //in the middle of a measurement so clear the MeasurementComplete flag
tWind2 = micros() - t0[3];
enableInterrupt(pinWind2, ISR_Wind2_RISING, RISING);
}        
    

void serialEvent3() 
// Nexus server's NMEA0183 data, 4800 baud, 8N1. 
// Use the server's NMEA Return and DC power ground to get the correct polarity (RS422 logic is inverted).
// Mega2560's Serial3 RX3 is on pin 15
{
  while (Serial3.available()) 
  {
    char inChar = (char)Serial3.read();    // get the new byte:
    inputString3 += inChar;                // append it to the inputString:
    if (inChar == '\n')                    // if the incoming character is a newline set the stringComplete flag so the main loop can service it
    {                 
      stringComplete3 = true;
    }
  }
}

void serialEvent2() 
// Garmin GPS data, 4800 baud, must invert the signal
// Mega2560's Serial2 RX2 is on pin 17
{
  while (Serial2.available()) 
  {
    char inChar = (char)Serial2.read();    // get the new byte:
    inputString2 += inChar;                // append it to the inputString:
    if (inChar == '\n')                    // if the incoming character is a newline set the stringComplete flag so the main loop can service it
    {                 
      stringComplete2 = true;
    }
  }
}
    